@@ -5,6 +5,7 @@
  * Copyright (C) 1998, 1999 Alex Roberts, Evan Lawrence
  * Copyright (C) 2000, 2001 Chema Celorio, Paolo Maggi
  * Copyright (C) 2002-2005 Paolo Maggi
+ * Copyright (C) 2014 Sébastien Wilmet
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -41,21 +42,15 @@
 #include "gedit-close-confirmation-dialog.h"
 
 #define GEDIT_OPEN_DIALOG_KEY "gedit-open-dialog-key"
-#define GEDIT_TAB_TO_SAVE_AS "gedit-tab-to-save-as"
-#define GEDIT_LIST_OF_TABS_TO_SAVE_AS "gedit-list-of-tabs-to-save-as"
 #define GEDIT_IS_CLOSING_ALL "gedit-is-closing-all"
 #define GEDIT_NOTEBOOK_TO_CLOSE "gedit-notebook-to-close"
 #define GEDIT_IS_QUITTING "gedit-is-quitting"
-#define GEDIT_IS_CLOSING_TAB "gedit-is-closing-tab"
 #define GEDIT_IS_QUITTING_ALL "gedit-is-quitting-all"
 
 static void tab_state_changed_while_saving (GeditTab    *tab,
 					    GParamSpec  *pspec,
 					    GeditWindow *window);
 
-static void save_as_tab (GeditTab    *tab,
-			 GeditWindow *window);
-
 void
 _gedit_cmd_file_new (GSimpleAction *action,
                      GVariant      *parameter,
@@ -705,144 +700,98 @@ get_compression_type_from_file (GFile *file)
 }
 
 static void
-save_finish_cb (GeditTab     *tab,
-		GAsyncResult *result,
-		gpointer      user_data)
+tab_save_as_ready_cb (GeditTab     *tab,
+		      GAsyncResult *result,
+		      GTask        *task)
 {
-	_gedit_tab_save_finish (tab, result);
+	gboolean success = _gedit_tab_save_finish (tab, result);
+	g_task_return_boolean (task, success);
+	g_object_unref (task);
 }
 
 static void
 save_dialog_response_cb (GeditFileChooserDialog *dialog,
-                         gint                    response_id,
-                         GeditWindow            *window)
+			 gint                    response_id,
+			 GTask                  *task)
 {
 	GeditTab *tab;
-	gpointer data;
-	GSList *tabs_to_save_as;
+	GeditWindow *window;
+	GeditDocument *doc;
+	GtkSourceFile *file;
+	GFile *location;
+	gchar *parse_name;
+	GtkSourceNewlineType newline_type;
+	GtkSourceCompressionType compression_type;
+	GtkSourceCompressionType current_compression_type;
+	const GtkSourceEncoding *encoding;
 
 	gedit_debug (DEBUG_COMMANDS);
 
-	tab = GEDIT_TAB (g_object_get_data (G_OBJECT (dialog),
-					    GEDIT_TAB_TO_SAVE_AS));
+	tab = g_task_get_source_object (task);
+	window = g_task_get_task_data (task);
 
 	if (response_id != GTK_RESPONSE_OK)
 	{
 		gedit_file_chooser_dialog_destroy (dialog);
-
-		goto save_next_tab;
-	}
-
-	if (tab != NULL)
-	{
-		GFile *location;
-		GeditDocument *doc;
-		GtkSourceFile *file;
-		gchar *parse_name;
-		GtkSourceNewlineType newline_type;
-		GtkSourceCompressionType compression_type;
-		GtkSourceCompressionType current_compression_type;
-		const GtkSourceEncoding *encoding;
-
-		doc = gedit_tab_get_document (tab);
-		file = gedit_document_get_file (doc);
-
-		location = gedit_file_chooser_dialog_get_file (dialog);
-		g_return_if_fail (location != NULL);
-
-		compression_type = get_compression_type_from_file (location);
-		current_compression_type = gtk_source_file_get_compression_type (file);
-
-		if ((compression_type == GTK_SOURCE_COMPRESSION_TYPE_NONE) !=
-		    (current_compression_type == GTK_SOURCE_COMPRESSION_TYPE_NONE))
-		{
-			GtkWindow *dialog_window = gedit_file_chooser_dialog_get_window (dialog);
-
-			if (!change_compression (dialog_window,
-			                         location,
-			                         compression_type != GTK_SOURCE_COMPRESSION_TYPE_NONE))
-			{
-				gedit_file_chooser_dialog_destroy (dialog);
-				g_object_unref (location);
-
-				goto save_next_tab;
-			}
-		}
-
-		encoding = gedit_file_chooser_dialog_get_encoding (dialog);
-		newline_type = gedit_file_chooser_dialog_get_newline_type (dialog);
-
-		gedit_file_chooser_dialog_destroy (dialog);
-
-		doc = gedit_tab_get_document (tab);
-		g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
-
-		parse_name = g_file_get_parse_name (location);
-
-		gedit_statusbar_flash_message (GEDIT_STATUSBAR (window->priv->statusbar),
-					        window->priv->generic_message_cid,
-					       _("Saving file '%s'\342\200\246"),
-					       parse_name);
-
-		g_free (parse_name);
-
-		/* Let's remember the dir we navigated to, even if the saving fails... */
-		_gedit_window_set_default_location (window, location);
-
-		_gedit_tab_save_as_async (tab,
-					  location,
-					  encoding,
-					  newline_type,
-					  compression_type,
-					  NULL,
-					  (GAsyncReadyCallback) save_finish_cb,
-					  NULL);
-
-		g_object_unref (location);
-	}
-
-save_next_tab:
-
-	data = g_object_get_data (G_OBJECT (window),
-				  GEDIT_LIST_OF_TABS_TO_SAVE_AS);
-	if (data == NULL)
-	{
+		g_task_return_boolean (task, FALSE);
+		g_object_unref (task);
 		return;
 	}
 
-	/* Save As the next tab of the list (we are Saving All files) */
-	tabs_to_save_as = (GSList *)data;
-	g_return_if_fail (tab == GEDIT_TAB (tabs_to_save_as->data));
+	doc = gedit_tab_get_document (tab);
+	file = gedit_document_get_file (doc);
 
-	/* Remove the first item of the list */
-	tabs_to_save_as = g_slist_delete_link (tabs_to_save_as,
-					       tabs_to_save_as);
+	location = gedit_file_chooser_dialog_get_file (dialog);
+	g_return_if_fail (location != NULL);
 
-	g_object_set_data (G_OBJECT (window),
-			   GEDIT_LIST_OF_TABS_TO_SAVE_AS,
-			   tabs_to_save_as);
+	compression_type = get_compression_type_from_file (location);
+	current_compression_type = gtk_source_file_get_compression_type (file);
 
-	if (tabs_to_save_as != NULL)
+	if ((compression_type == GTK_SOURCE_COMPRESSION_TYPE_NONE) !=
+	    (current_compression_type == GTK_SOURCE_COMPRESSION_TYPE_NONE))
 	{
-		tab = GEDIT_TAB (tabs_to_save_as->data);
+		GtkWindow *dialog_window = gedit_file_chooser_dialog_get_window (dialog);
 
-		if (GPOINTER_TO_BOOLEAN (g_object_get_data (G_OBJECT (tab),
-							    GEDIT_IS_CLOSING_TAB)))
+		if (!change_compression (dialog_window,
+					 location,
+					 compression_type != GTK_SOURCE_COMPRESSION_TYPE_NONE))
 		{
-			g_object_set_data (G_OBJECT (tab),
-					   GEDIT_IS_CLOSING_TAB,
-					   NULL);
+			gedit_file_chooser_dialog_destroy (dialog);
+			g_object_unref (location);
 
-			/* Trace tab state changes */
-			g_signal_connect (tab,
-					  "notify::state",
-					  G_CALLBACK (tab_state_changed_while_saving),
-					  window);
+			g_task_return_boolean (task, FALSE);
+			g_object_unref (task);
+			return;
 		}
-
-		gedit_window_set_active_tab (window, tab);
-		save_as_tab (tab, window);
 	}
+
+	encoding = gedit_file_chooser_dialog_get_encoding (dialog);
+	newline_type = gedit_file_chooser_dialog_get_newline_type (dialog);
+
+	gedit_file_chooser_dialog_destroy (dialog);
+
+	parse_name = g_file_get_parse_name (location);
+
+	gedit_statusbar_flash_message (GEDIT_STATUSBAR (window->priv->statusbar),
+				       window->priv->generic_message_cid,
+				       _("Saving file '%s'\342\200\246"),
+				       parse_name);
+
+	g_free (parse_name);
+
+	/* Let's remember the dir we navigated to, even if the saving fails... */
+	_gedit_window_set_default_location (window, location);
+
+	_gedit_tab_save_as_async (tab,
+				  location,
+				  encoding,
+				  newline_type,
+				  compression_type,
+				  g_task_get_cancellable (task),
+				  (GAsyncReadyCallback) tab_save_as_ready_cb,
+				  task);
+
+	g_object_unref (location);
 }
 
 static GtkFileChooserConfirmation
@@ -882,10 +831,15 @@ confirm_overwrite_callback (GeditFileChooserDialog *dialog,
 	return res;
 }
 
+/* Call save_as_tab_finish() in @callback. */
 static void
-save_as_tab (GeditTab    *tab,
-	     GeditWindow *window)
+save_as_tab_async (GeditTab            *tab,
+		   GeditWindow         *window,
+		   GCancellable        *cancellable,
+		   GAsyncReadyCallback  callback,
+		   gpointer             user_data)
 {
+	GTask *task;
 	GeditFileChooserDialog *save_dialog;
 	GtkWindowGroup *window_group;
 	GtkWindow *dialog_window;
@@ -900,6 +854,9 @@ save_as_tab (GeditTab    *tab,
 
 	gedit_debug (DEBUG_COMMANDS);
 
+	task = g_task_new (tab, cancellable, callback, user_data);
+	g_task_set_task_data (task, g_object_ref (window), g_object_unref);
+
 	save_dialog = gedit_file_chooser_dialog_create (_("Save As"),
 							GTK_WINDOW (window),
 							GEDIT_FILE_CHOOSER_SAVE |
@@ -960,7 +917,7 @@ save_as_tab (GeditTab    *tab,
 		g_free (docname);
 	}
 
-	/* Set suggested encoding */
+	/* Set suggested encoding and newline type. */
 	encoding = gtk_source_file_get_encoding (file);
 
 	if (encoding == NULL)
@@ -976,53 +933,154 @@ save_as_tab (GeditTab    *tab,
 	gedit_file_chooser_dialog_set_newline_type (GEDIT_FILE_CHOOSER_DIALOG (save_dialog),
 	                                            newline_type);
 
-	g_object_set_data (G_OBJECT (save_dialog),
-			   GEDIT_TAB_TO_SAVE_AS,
-			   tab);
-
 	g_signal_connect (save_dialog,
 			  "response",
 			  G_CALLBACK (save_dialog_response_cb),
-			  window);
+			  task);
 
 	gedit_file_chooser_dialog_show (save_dialog);
 }
 
-static void
-save_tab (GeditTab    *tab,
-	  GeditWindow *window)
+static gboolean
+save_as_tab_finish (GeditTab     *tab,
+		    GAsyncResult *result)
 {
-	GeditDocument *doc;
+	g_return_val_if_fail (g_task_is_valid (result, tab), FALSE);
+
+	return g_task_propagate_boolean (G_TASK (result), NULL);
+}
+
+static void
+save_as_tab_ready_cb (GeditTab     *tab,
+		      GAsyncResult *result,
+		      GTask        *task)
+{
+	gboolean success = save_as_tab_finish (tab, result);
+
+	g_task_return_boolean (task, success);
+	g_object_unref (task);
+}
+
+static void
+tab_save_ready_cb (GeditTab     *tab,
+		   GAsyncResult *result,
+		   GTask        *task)
+{
+	gboolean success = _gedit_tab_save_finish (tab, result);
+
+	g_task_return_boolean (task, success);
+	g_object_unref (task);
+}
+
+/**
+ * gedit_commands_save_document_async:
+ * @document: the #GeditDocument to save.
+ * @window: a #GeditWindow.
+ * @cancellable: (nullable): optional #GCancellable object, %NULL to ignore.
+ * @callback: (scope async): a #GAsyncReadyCallback to call when the operation
+ *   is finished.
+ * @user_data: (closure): the data to pass to the @callback function.
+ *
+ * Asynchronously save the @document. @document must belong to @window. The
+ * source object of the async task is @document (which will be the first
+ * parameter of the #GAsyncReadyCallback).
+ *
+ * When the operation is finished, @callback will be called. You can then call
+ * gedit_commands_save_document_finish() to get the result of the operation.
+ *
+ * Since: 3.14
+ */
+void
+gedit_commands_save_document_async (GeditDocument       *document,
+				    GeditWindow         *window,
+				    GCancellable        *cancellable,
+				    GAsyncReadyCallback  callback,
+				    gpointer             user_data)
+{
+	GTask *task;
+	GeditTab *tab;
 	gchar *uri_for_display;
 
 	gedit_debug (DEBUG_COMMANDS);
 
-	g_return_if_fail (GEDIT_IS_TAB (tab));
+	g_return_if_fail (GEDIT_IS_DOCUMENT (document));
 	g_return_if_fail (GEDIT_IS_WINDOW (window));
+	g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
 
-	doc = gedit_tab_get_document (tab);
-	g_return_if_fail (GEDIT_IS_DOCUMENT (doc));
+	task = g_task_new (document, cancellable, callback, user_data);
 
-	if (gedit_document_is_untitled (doc) ||
-	    gedit_document_get_readonly (doc))
+	tab = gedit_tab_get_from_document (document);
+
+	if (gedit_document_is_untitled (document) ||
+	    gedit_document_get_readonly (document))
 	{
 		gedit_debug_message (DEBUG_COMMANDS, "Untitled or Readonly");
-		save_as_tab (tab, window);
+
+		save_as_tab_async (tab,
+				   window,
+				   cancellable,
+				   (GAsyncReadyCallback) save_as_tab_ready_cb,
+				   task);
 		return;
 	}
 
-	uri_for_display = gedit_document_get_uri_for_display (doc);
+	uri_for_display = gedit_document_get_uri_for_display (document);
 	gedit_statusbar_flash_message (GEDIT_STATUSBAR (window->priv->statusbar),
-				        window->priv->generic_message_cid,
+				       window->priv->generic_message_cid,
 				       _("Saving file '%s'\342\200\246"),
 				       uri_for_display);
 
 	g_free (uri_for_display);
 
 	_gedit_tab_save_async (tab,
-			       NULL,
-			       (GAsyncReadyCallback) save_finish_cb,
-			       NULL);
+			       cancellable,
+			       (GAsyncReadyCallback) tab_save_ready_cb,
+			       task);
+}
+
+/**
+ * gedit_commands_save_document_finish:
+ * @document: a #GeditDocument.
+ * @result: a #GAsyncResult.
+ *
+ * Finishes an asynchronous document saving operation started with
+ * gedit_commands_save_document_async().
+ *
+ * Note that there is no error parameter because the errors are already handled
+ * by gedit.
+ *
+ * Returns: %TRUE if the document has been correctly saved, %FALSE otherwise.
+ * Since: 3.14
+ */
+gboolean
+gedit_commands_save_document_finish (GeditDocument *document,
+				     GAsyncResult  *result)
+{
+	g_return_val_if_fail (g_task_is_valid (result, document), FALSE);
+
+	return g_task_propagate_boolean (G_TASK (result), NULL);
+}
+
+static void
+save_tab_ready_cb (GeditDocument *doc,
+		   GAsyncResult  *result,
+		   gpointer       user_data)
+{
+	gedit_commands_save_document_finish (doc, result);
+}
+
+/* Save tab asynchronously, but without results. */
+static void
+save_tab (GeditTab    *tab,
+	  GeditWindow *window)
+{
+	GeditDocument *doc = gedit_tab_get_document (tab);
+
+	gedit_commands_save_document_async (doc,
+					    window,
+					    NULL,
+					    (GAsyncReadyCallback) save_tab_ready_cb,
+					    NULL);
 }
 
 void
@@ -1042,6 +1100,14 @@ _gedit_cmd_file_save (GSimpleAction *action,
 	}
 }
 
+static void
+_gedit_cmd_file_save_as_cb (GeditTab     *tab,
+			    GAsyncResult *result,
+			    gpointer      user_data)
+{
+	save_as_tab_finish (tab, result);
+}
+
 void
 _gedit_cmd_file_save_as (GSimpleAction *action,
                          GVariant      *parameter,
@@ -1055,10 +1121,153 @@ _gedit_cmd_file_save_as (GSimpleAction *action,
 	tab = gedit_window_get_active_tab (window);
 	if (tab != NULL)
 	{
-		save_as_tab (tab, window);
+		save_as_tab_async (tab,
+				   window,
+				   NULL,
+				   (GAsyncReadyCallback) _gedit_cmd_file_save_as_cb,
+				   NULL);
 	}
 }
 
+static void
+quit_if_needed (GeditWindow *window)
+{
+	gboolean is_quitting;
+	gboolean is_quitting_all;
+
+	is_quitting = GPOINTER_TO_BOOLEAN (g_object_get_data (G_OBJECT (window),
+							      GEDIT_IS_QUITTING));
+
+	is_quitting_all = GPOINTER_TO_BOOLEAN (g_object_get_data (G_OBJECT (window),
+							          GEDIT_IS_QUITTING_ALL));
+
+	if (is_quitting)
+	{
+		gtk_widget_destroy (GTK_WIDGET (window));
+	}
+
+	if (is_quitting_all)
+	{
+		GtkApplication *app;
+
+		app = GTK_APPLICATION (g_application_get_default ());
+
+		if (gtk_application_get_windows (app) == NULL)
+		{
+			g_application_quit (G_APPLICATION (app));
+		}
+	}
+}
+
+static gboolean
+really_close_tab (GeditTab *tab)
+{
+	GtkWidget *toplevel;
+	GeditWindow *window;
+
+	gedit_debug (DEBUG_COMMANDS);
+
+	g_return_val_if_fail (gedit_tab_get_state (tab) == GEDIT_TAB_STATE_CLOSING,
+			      FALSE);
+
+	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (tab));
+	g_return_val_if_fail (GEDIT_IS_WINDOW (toplevel), FALSE);
+
+	window = GEDIT_WINDOW (toplevel);
+
+	gedit_window_close_tab (window, tab);
+
+	if (gedit_window_get_active_tab (window) == NULL)
+	{
+		quit_if_needed (window);
+	}
+
+	return FALSE;
+}
+
+static void
+close_tab (GeditTab *tab)
+{
+	GeditDocument *doc;
+
+	doc = gedit_tab_get_document (tab);
+	g_return_if_fail (doc != NULL);
+
+	/* If the user has modified again the document, do not close the tab. */
+	if (_gedit_document_needs_saving (doc))
+	{
+		return;
+	}
+
+	/* Close the document only if it has been succesfully saved.
+	 * Tab state is set to CLOSING (it is a state without exiting
+	 * transitions) and the tab is closed in an idle handler.
+	 */
+	_gedit_tab_mark_for_closing (tab);
+
+	g_idle_add_full (G_PRIORITY_HIGH_IDLE,
+			 (GSourceFunc) really_close_tab,
+			 tab,
+			 NULL);
+}
+
+typedef struct _SaveAsData SaveAsData;
+
+struct _SaveAsData
+{
+	/* Reffed */
+	GeditWindow *window;
+
+	/* List of reffed GeditTab's */
+	GSList *tabs_to_save_as;
+
+	guint close_tabs : 1;
+};
+
+static void save_as_documents_list (SaveAsData *data);
+
+static void
+save_as_documents_list_cb (GeditTab     *tab,
+			   GAsyncResult *result,
+			   SaveAsData   *data)
+{
+	gboolean saved = save_as_tab_finish (tab, result);
+
+	if (saved && data->close_tabs)
+	{
+		close_tab (tab);
+	}
+
+	g_return_if_fail (tab == GEDIT_TAB (data->tabs_to_save_as->data));
+	g_object_unref (data->tabs_to_save_as->data);
+	data->tabs_to_save_as = g_slist_delete_link (data->tabs_to_save_as,
+						     data->tabs_to_save_as);
+
+	if (data->tabs_to_save_as != NULL)
+	{
+		save_as_documents_list (data);
+	}
+	else
+	{
+		g_object_unref (data->window);
+		g_slice_free (SaveAsData, data);
+	}
+}
+
+static void
+save_as_documents_list (SaveAsData *data)
+{
+	GeditTab *next_tab = GEDIT_TAB (data->tabs_to_save_as->data);
+
+	gedit_window_set_active_tab (data->window, next_tab);
+
+	save_as_tab_async (next_tab,
+			   data->window,
+			   NULL,
+			   (GAsyncReadyCallback) save_as_documents_list_cb,
+			   data);
+}
+
 /*
  * The docs in the list must belong to the same GeditWindow.
  */
@@ -1066,8 +1275,8 @@ static void
 save_documents_list (GeditWindow *window,
 		     GList       *docs)
 {
+	SaveAsData *data = NULL;
 	GList *l;
-	GSList *tabs_to_save_as = NULL;
 
 	gedit_debug (DEBUG_COMMANDS);
 
@@ -1100,7 +1309,16 @@ save_documents_list (GeditWindow *window,
 				if (gedit_document_is_untitled (doc) ||
 				    gedit_document_get_readonly (doc))
 				{
-					tabs_to_save_as = g_slist_prepend (tabs_to_save_as, tab);
+					if (data == NULL)
+					{
+						data = g_slice_new (SaveAsData);
+						data->window = g_object_ref (window);
+						data->tabs_to_save_as = NULL;
+						data->close_tabs = FALSE;
+					}
+
+					data->tabs_to_save_as = g_slist_prepend (data->tabs_to_save_as,
+										 g_object_ref (tab));
 				}
 				else
 				{
@@ -1142,26 +1360,21 @@ save_documents_list (GeditWindow *window,
 		}
 	}
 
-	if (tabs_to_save_as != NULL)
+	if (data != NULL)
 	{
-		GeditTab *tab;
-
-		tabs_to_save_as = g_slist_reverse (tabs_to_save_as);
-
-		g_return_if_fail (g_object_get_data (G_OBJECT (window),
-						     GEDIT_LIST_OF_TABS_TO_SAVE_AS) == NULL);
-
-		g_object_set_data (G_OBJECT (window),
-				   GEDIT_LIST_OF_TABS_TO_SAVE_AS,
-				   tabs_to_save_as);
-
-		tab = GEDIT_TAB (tabs_to_save_as->data);
-
-		gedit_window_set_active_tab (window, tab);
-		save_as_tab (tab, window);
+		data->tabs_to_save_as = g_slist_reverse (data->tabs_to_save_as);
+		save_as_documents_list (data);
 	}
 }
 
+/**
+ * gedit_commands_save_all_documents:
+ * @window: a #GeditWindow.
+ *
+ * Asynchronously save all documents belonging to @window. The result of the
+ * operation is not available, so it's difficult to know whether all the
+ * documents are correctly saved.
+ */
 void
 gedit_commands_save_all_documents (GeditWindow *window)
 {
@@ -1186,6 +1399,14 @@ _gedit_cmd_file_save_all (GSimpleAction *action,
 	gedit_commands_save_all_documents (GEDIT_WINDOW (user_data));
 }
 
+/**
+ * gedit_commands_save_document:
+ * @window: a #GeditWindow.
+ * @document: the #GeditDocument to save.
+ *
+ * Asynchronously save @document. @document must belong to @window. If you need
+ * the result of the operation, use gedit_commands_save_document_async().
+ */
 void
 gedit_commands_save_document (GeditWindow   *window,
                               GeditDocument *document)
@@ -1412,63 +1633,6 @@ _gedit_cmd_file_revert (GSimpleAction *action,
 	gtk_widget_show (dialog);
 }
 
-static void
-quit_if_needed (GeditWindow *window)
-{
-	gboolean is_quitting;
-	gboolean is_quitting_all;
-
-	is_quitting = GPOINTER_TO_BOOLEAN (g_object_get_data (G_OBJECT (window),
-							      GEDIT_IS_QUITTING));
-
-	is_quitting_all = GPOINTER_TO_BOOLEAN (g_object_get_data (G_OBJECT (window),
-							          GEDIT_IS_QUITTING_ALL));
-
-	if (is_quitting)
-	{
-		gtk_widget_destroy (GTK_WIDGET (window));
-	}
-
-	if (is_quitting_all)
-	{
-		GtkApplication *app;
-
-		app = GTK_APPLICATION (g_application_get_default ());
-
-		if (gtk_application_get_windows (app) == NULL)
-		{
-			g_application_quit (G_APPLICATION (app));
-		}
-	}
-}
-
-/* Close tab */
-static gboolean
-really_close_tab (GeditTab *tab)
-{
-	GtkWidget *toplevel;
-	GeditWindow *window;
-
-	gedit_debug (DEBUG_COMMANDS);
-
-	g_return_val_if_fail (gedit_tab_get_state (tab) == GEDIT_TAB_STATE_CLOSING,
-			      FALSE);
-
-	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (tab));
-	g_return_val_if_fail (GEDIT_IS_WINDOW (toplevel), FALSE);
-
-	window = GEDIT_WINDOW (toplevel);
-
-	gedit_window_close_tab (window, tab);
-
-	if (gedit_window_get_active_tab (window) == NULL)
-	{
-		quit_if_needed (window);
-	}
-
-	return FALSE;
-}
-
 static void
 tab_state_changed_while_saving (GeditTab    *tab,
 				GParamSpec  *pspec,
@@ -1485,33 +1649,11 @@ tab_state_changed_while_saving (GeditTab    *tab,
 	 */
 	if (state == GEDIT_TAB_STATE_NORMAL)
 	{
-		GeditDocument *doc;
-
 		g_signal_handlers_disconnect_by_func (tab,
 						      G_CALLBACK (tab_state_changed_while_saving),
 					      	      window);
 
-		doc = gedit_tab_get_document (tab);
-		g_return_if_fail (doc != NULL);
-
-		/* If the saving operation failed or was interrupted, then the
-		 * document is still "modified" -> do not close the tab.
-		 */
-		if (_gedit_document_needs_saving (doc))
-		{
-			return;
-		}
-
-		/* Close the document only if it has been succesfully saved.
-		 * Tab state is set to CLOSING (it is a state without exiting
-		 * transitions) and the tab is closed in an idle handler.
-		 */
-		_gedit_tab_mark_for_closing (tab);
-
-		g_idle_add_full (G_PRIORITY_HIGH_IDLE,
-				 (GSourceFunc)really_close_tab,
-				 tab,
-				 NULL);
+		close_tab (tab);
 	}
 }
 
@@ -1530,26 +1672,6 @@ save_and_close (GeditTab    *tab,
 	save_tab (tab, window);
 }
 
-static void
-save_as_and_close (GeditTab    *tab,
-		   GeditWindow *window)
-{
-	gedit_debug (DEBUG_COMMANDS);
-
-	g_object_set_data (G_OBJECT (tab),
-			   GEDIT_IS_CLOSING_TAB,
-			   NULL);
-
-	/* Trace tab state changes */
-	g_signal_connect (tab,
-			  "notify::state",
-			  G_CALLBACK (tab_state_changed_while_saving),
-			  window);
-
-	gedit_window_set_active_tab (window, tab);
-	save_as_tab (tab, window);
-}
-
 static void
 save_and_close_documents (GList         *docs,
                           GeditWindow   *window,
@@ -1558,7 +1680,7 @@ save_and_close_documents (GList         *docs,
 	GList *tabs;
 	GList *l;
 	GSList *sl;
-	GSList *tabs_to_save_as = NULL;
+	SaveAsData *data = NULL;
 	GSList *tabs_to_save_and_close = NULL;
 	GList *tabs_to_close = NULL;
 
@@ -1630,11 +1752,16 @@ save_and_close_documents (GList         *docs,
 				if (gedit_document_is_untitled (doc) ||
 				    gedit_document_get_readonly (doc))
 				{
-					g_object_set_data (G_OBJECT (tab),
-							   GEDIT_IS_CLOSING_TAB,
-							   GBOOLEAN_TO_POINTER (TRUE));
+					if (data == NULL)
+					{
+						data = g_slice_new (SaveAsData);
+						data->window = g_object_ref (window);
+						data->tabs_to_save_as = NULL;
+						data->close_tabs = TRUE;
+					}
 
-					tabs_to_save_as = g_slist_prepend (tabs_to_save_as, tab);
+					data->tabs_to_save_as = g_slist_prepend (data->tabs_to_save_as,
+										 g_object_ref (tab));
 				}
 				else
 				{
@@ -1663,23 +1790,11 @@ save_and_close_documents (GList         *docs,
 
 	g_slist_free (tabs_to_save_and_close);
 
-	/* Save As and close all the files in tabs_to_save_as  */
-	if (tabs_to_save_as != NULL)
+	/* Save As and close all the files in data->tabs_to_save_as. */
+	if (data != NULL)
 	{
-		GeditTab *tab;
-
-		tabs_to_save_as = g_slist_reverse (tabs_to_save_as);
-
-		g_return_if_fail (g_object_get_data (G_OBJECT (window),
-						     GEDIT_LIST_OF_TABS_TO_SAVE_AS) == NULL);
-
-		g_object_set_data (G_OBJECT (window),
-				   GEDIT_LIST_OF_TABS_TO_SAVE_AS,
-				   tabs_to_save_as);
-
-		tab = GEDIT_TAB (tabs_to_save_as->data);
-
-		save_as_and_close (tab, window);
+		data->tabs_to_save_as = g_slist_reverse (data->tabs_to_save_as);
+		save_as_documents_list (data);
 	}
 }
 
