@@ -1662,17 +1662,11 @@ gedit_document_set_metadata (GeditDocument *doc,
 		else
 		{
 			/* Unset the key */
-			g_file_info_set_attribute (info,
-						   key,
-						   G_FILE_ATTRIBUTE_TYPE_INVALID,
-						   NULL);
+			g_file_info_remove_attribute (info, key);
 
 			if (doc->priv->metadata_info != NULL)
 			{
-				g_file_info_set_attribute (doc->priv->metadata_info,
-							   key,
-							   G_FILE_ATTRIBUTE_TYPE_INVALID,
-							   NULL);
+				g_file_info_remove_attribute (doc->priv->metadata_info, key);
 			}
 		}
 	}
