@@ -310,7 +310,10 @@ set_click_policy_property (GeditFileBrowserView            *obj,
 	if (click_policy == GEDIT_FILE_BROWSER_VIEW_CLICK_POLICY_SINGLE)
 	{
 		if (obj->priv->hand_cursor == NULL)
-			obj->priv->hand_cursor = gdk_cursor_new (GDK_HAND2);
+		{
+			display = gtk_widget_get_display (GTK_WIDGET (obj));
+			obj->priv->hand_cursor = gdk_cursor_new_for_display (display, GDK_HAND2);
+		}
 	}
 	else if (click_policy == GEDIT_FILE_BROWSER_VIEW_CLICK_POLICY_DOUBLE)
 	{
